# Copyright 2020 Andrzej Cichocki

# This file is part of soak.
#
# soak is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# soak is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with soak.  If not, see <http://www.gnu.org/licenses/>.

'Process aridity templates as per all soak.arid configs in directory tree.'
from . import cpuexecutor
from .context import createparent
from .multifork import Tasks
from .terminal import getterminal
from argparse import ArgumentParser
from diapyr.util import invokeall
from functools import partial
from lagoon import diff
from lagoon.util import atomic
from pathlib import Path
import logging, os, sys

log = logging.getLogger(__name__)

class SoakConfig:

    soakkey = 'soak'

    def __init__(self, parent, configpath):
        self.node = (-parent).childctrl().node
        self.node.cwd = str(configpath.parent.resolve())
        (-self.node).printf(". %s", configpath.name)
        self.reltargets = [Path(rt) for rt, _ in (-getattr(self.node, self.soakkey)).scope().resolvables.items()]
        self.dirpath = configpath.parent

    def process(self, reltarget):
        with atomic(self.dirpath / reltarget) as partpath:
            (-self.node).scope().resolved(self.soakkey, str(reltarget), 'data').writeout(partpath)

    def origtext(self, reltarget):
        return getattr(getattr(self.node, self.soakkey), str(reltarget)).diff

    def diff(self, origtext, reltarget):
        diff._us[print]('--color=always', '-', self.dirpath / reltarget, input = origtext, check = False)

def main():
    logging.basicConfig(format = "[%(levelname)s] %(message)s", level = logging.DEBUG)
    parser = ArgumentParser()
    parser.add_argument('-n', action = 'store_true')
    parser.add_argument('-d', action = 'store_true')
    parser.add_argument('-v', action = 'store_true')
    config = parser.parse_args()
    if not config.v:
        logging.getLogger().setLevel(logging.INFO)
    soakroot = Path('.')
    parent = createparent(soakroot)
    soakconfigs = [SoakConfig(parent, p) for p in soakroot.rglob('soak.arid')]
    if not config.n:
        terminal = getterminal()
        tasks = Tasks()
        for soakconfig in soakconfigs:
            for reltarget in soakconfig.reltargets:
                task = partial(soakconfig.process, reltarget)
                task.index = len(tasks)
                task.target = soakconfig.dirpath / reltarget
                terminal.head(task.index, task.target, dark = True)
                tasks.append(task)
        tasks.started = lambda task: terminal.head(task.index, task.target, rev = True)
        tasks.stdout = lambda task, line: terminal.log(task.index, sys.stdout, line)
        tasks.stderr = lambda task, line: terminal.log(task.index, sys.stderr, line)
        tasks.stopped = lambda task: terminal.head(task.index, task.target)
        tasks.drain(os.cpu_count())
    if config.d:
        with cpuexecutor() as executor:
            diffs = []
            for soakconfig in soakconfigs:
                for reltarget in soakconfig.reltargets:
                    diffs.append(partial(lambda soakconfig, origtextfuture, reltarget: soakconfig.diff(origtextfuture.result(), reltarget),
                            soakconfig, executor.submit(soakconfig.origtext, reltarget), reltarget))
            invokeall(diffs)

if '__main__' == __name__:
    main()



# Copyright 2020 Andrzej Cichocki

# This file is part of soak.
#
# soak is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# soak is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with soak.  If not, see <http://www.gnu.org/licenses/>.

from diapyr.util import singleton
from itertools import islice
from lagoon import tput
from subprocess import CalledProcessError
import sys

class AbstractLog:

    def head(self, index, obj, rev = False, dark = False):
        return self.headimpl(index, obj, rev, dark)

class Terminal(AbstractLog):

    class Section:

        height = 0

    def __init__(self, width):
        self.sections = []
        self.width = width

    def _common(self, index, tonewh):
        dy = sum(s.height for s in islice(self.sections, index + 1, None))
        section = self.sections[index]
        oldh = section.height
        section.height = newh = tonewh(oldh)
        if dy:
            tput.cuu(dy, stdout = sys.stderr)
        if newh > oldh:
            tput.il(newh - oldh, stdout = sys.stderr)
        return dy, oldh, newh

    def headimpl(self, index, obj, rev, dark):
        for _ in range(len(self.sections), index + 1):
            self.sections.append(self.Section())
        dy, oldh, newh = self._common(index, lambda h: max(1, h))
        if oldh:
            tput.cuu(oldh, stdout = sys.stderr)
        if rev:
            tput.rev(stdout = sys.stderr)
        if dark:
            tput.setaf(0, stdout = sys.stderr)
        sys.stderr.write(f"[{obj}]{tput.sgr0()}\n")
        sys.stderr.write('\n' * (newh - 1 + dy))

    def log(self, index, stream, line):
        dy, oldh, newh = self._common(index, lambda h: h + 1)
        noeol, = line.splitlines()
        eol = line[len(noeol):]
        chunks = [noeol[i:i + self.width] for i in range(0, len(noeol), self.width)]
        stream.write(chunks[0])
        for c in islice(chunks, 1, None):
            stream.flush()
            tput.hpa(0, stdout = sys.stderr)
            sys.stderr.flush()
            stream.write(c)
        if eol:
            stream.write(eol)
        else:
            stream.flush()
        sys.stderr.write('\n' * ((not eol) + dy))

@singleton
class LogFile(AbstractLog):

    def headimpl(self, index, obj, rev, dark):
        if not dark:
            print('Damp:' if rev else 'Soaked:', obj, file = sys.stderr)

    def log(self, index, stream, line):
        stream.write(line)

def getterminal():
    try:
        width = int(tput.cols())
    except CalledProcessError:
        return LogFile
    return Terminal(width)



# Copyright 2020 Andrzej Cichocki

# This file is part of soak.
#
# soak is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# soak is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with soak.  If not, see <http://www.gnu.org/licenses/>.

from base64 import b64decode, b64encode
from diapyr.util import invokeall
from select import select
import os, pickle, sys

class GoodResult:

    def __init__(self, value):
        self.value = value

    def get(self):
        return self.value

class BadResult:

    def __init__(self, exception):
        self.exception = exception

    def get(self):
        raise self.exception

class Job:

    ttl = 3

    def __init__(self, index, task):
        self.index = index
        self.task = task

    def start(self):
        r1, w1 = os.pipe()
        r2, w2 = os.pipe()
        rx, wx = os.pipe()
        pid = os.fork()
        if pid:
            os.close(w1)
            os.close(w2)
            os.close(wx)
            self.pid = pid
            return map(os.fdopen, [r1, r2, rx])
        os.close(r1)
        os.close(r2)
        os.close(rx)
        os.dup2(w1, 1)
        os.close(w1)
        os.dup2(w2, 2)
        os.close(w2)
        try:
            obj = GoodResult(self.task())
        except BaseException as e:
            obj = BadResult(e)
        os.write(wx, b64encode(pickle.dumps([self.index, obj])))
        sys.exit()

    def decr(self):
        self.ttl = ttl = self.ttl - 1
        if not ttl:
            os.waitpid(self.pid, 0)
            return True

class Tasks(list):

    def drain(self, limit):
        def report(task, line):
            index, obj = pickle.loads(b64decode(line))
            results[index] = obj.get
        streams = {}
        running = 0
        results = [None] * len(self)
        while self or streams:
            while self and running < limit:
                job = Job(len(results) - len(self), self.pop(0))
                r1, r2, rx = job.start()
                streams[r1] = job, self.stdout
                streams[r2] = job, self.stderr
                streams[rx] = job, report
                running += 1
                self.started(job.task)
            for r in select(streams, [], [])[0]:
                line = r.readline()
                if line:
                    job, callback = streams[r]
                    callback(job.task, line)
                else:
                    job = streams.pop(r)[0]
                    r.close()
                    if job.decr():
                        running -= 1
                        self.stopped(job.task)
        return invokeall(results)

    def started(self, task):
        pass

    def stdout(self, task, line):
        pass

    def stderr(self, task, line):
        pass

    def stopped(self, task):
        pass
